/*
 * emacs.e -- LaTeX indentation and line reflowing routines
 *
 * Copyright (C) 1997 Juan Jos\'e Garc\'ia Ripoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Comments:
 *
 * These routines are inspired in several algorithms from the AucTeX
 * package, but they have been heavily modified and improved to run
 * under EPM-6.03.
 *
 * You should have received an `emacs.txt' document explaining how
 * all this stuff works.
 *
 * To-do:
 *	+ Improve the logic (this relies on me learning more LaTeX)
 *	+ Add code for use of INDENT and FILLPAR in button bars.
 *	+ Maybe fixing FILLPAR to find the begining of the paragraph
 *	  as well as the end.
 *
 * The author: 	Juan Jose Garcia Ripoll <worm@arrakis.es>
 */

/*
 * TeX and LaTeX constants.
 */

const
-- Set to 1 if you want comments to be taken out in separate lines
COMMENTS_OUT_OF_LINES=1

-- Ammount of indentation introduced by \begin and '{'
LaTeX_indent_level = 2

-- Ammount of indentation introduced by \left
LaTeX_left_right_indent_level = LaTeX_indent_level

-- Ammount of indentation added by \item (usually negative)
LaTeX_item_indent = -LaTeX_indent_level

-- Matching LaTeX expressions that should go in their own lines
LaTeX_paragraph_commands = '\[|\]|'||
                           '(begin|end|part|chapter|label|'||
                           'caption|section|subsection|subsubsection|'||
                           'par|noindent|paragraph|include|'||
                           'includeonly|tableofcontents|appendix)~:a'

-- Escape sequence to find a paragraph start
LaTeX_paragraph_start = '(^%.*$|^.*[^\\]%.*$|^:o$|'||
                         '^:o\\('LaTeX_paragraph_commands'|item~:a)|'||
                         '^:o\$\$)'

/************************************************
 *	General EMACS-like procedures		*
 ************************************************/
--
-- Returns true if the SUB string is at the beginning
-- of STR. As we use LeftStr(), trailing spaces in
-- SUB are not taken into account.
--
defproc prefix(sub, str)
   return (LeftStr(str,length(sub)) == sub)

--
-- Returns the appropiate string to indent a line a fixed
-- amount. This uses tabs, which is the appropiate for
-- saving space but may confuse EPM. To solve this, use
--	return copies(' ', amount)
-- instead.
--
defproc indent_str(amount)
   return copies(\t, amount%8)||copies(' ', amount//8)

--
-- Removes all leading spaces and tabs from a text
--
defproc strip_indent(lin)
   return strip(strip(lin, 'L', ' '), 'L', \t)

--
-- Removes all leading spaces and tabs from text,
-- substracting from `col' the amount of characters
-- removed
--
defproc strip_indent_col(lin, col_old, var col_new)
   tmp = strip(strip(lin, 'L', ' '), 'L', \t)
   col_new = col_old + Length(tmp) - Length(lin)
   return tmp

--
-- Returns the effective indentation of the current line.
-- Every `tab' counts for 8 characteres (Isn't this the
-- standard in EPM for the 09x character?)
--
defproc current_indentation
   getline lin
   column = 0;
   do i = 1 to Length(lin)
      c = SubStr(lin,i,1)
      if c == ' ' then
         column = column + 1
      elseif c == \t then
         column = column + 8 - (column // 8)
      else
         return column
      endif
   enddo
   return 0

defproc where_to_split(max, line)
   n = LastPos(' ',LeftStr(line,max))
   if n = 0 then
      n = Pos(' ',line)
      if n = 0 then
         n = max
      endif
   endif
   return n

--
-- Splits a line without leading indentation characters (TEXT)
-- into two pieces (LINE and TEXT) so that the first one (LINE)
-- fits in a margin that is (MAX) characters wide.
--
-- Outputs:     1 if the line was successfully splitted without
--                further intervention
--              2 if the line was splitted and the second piece
--                was transformed into a standalone comment
--              0 if the line needed not to be splitted                
--
defproc split_line(max, var line, var text)
   if Length(text) <= max then
      line = text
      text = ''
      return 0
   endif
   line = LeftStr(text, max)
   n = LastPos(' ',line)
   if n == 0 then
      -- FIXME! If we get here then the user has written a word
      -- so long that the line cannot be safely splitted. Maybe
      -- we should issue a warning?
      n = max
   endif
   --
   -- Lines beginning with a comment are splitted into two
   -- comment lines
   --
   if LeftStr(text,1) == '%' then
      line = LeftStr(text, n)
      text = '%'SubStr(text,n+1)
      return 2
   endif
   nc = Pos('[^\\]%',text,1,'G')
compile if COMMENTS_OUT_OF_LINES=1
   --
   -- In the next line, comments inside sentences are taken out into
   -- separate lines.
   --
   if (nc > 0) & (nc < n) then
      n = nc
   endif
compile endif
   line = LeftStr(text, n)
   text = SubStr(text, n+1)
   return 1

--
--Reindents the current line to AMT columns
--
--FIXME! This indentation algorithm is not very smart. For instance,
--when using margins it doesn't split the line. And when the line
--plus the indentation is too long wierd things may occur
--
/*
defproc reindent_to(amt)
   getline line
   replaceline indent_str(amt)||line
*/

/************************************************
 *	General AucTeX-like procedures		*
 ************************************************/

--
-- Removes TeX comments from the end of LIN
--
defproc strip_comment(lin)
   n = Pos('[^\\]%',lin,1,'G')
   if n = 0 then
      if LeftStr(lin,1) == '%' then
         return ''
      else
         return lin
      endif
   endif
   return LeftStr(lin, n)

/*
defproc LaTeX_curr_env
   lastlin = .line
   lastcol = .col
   'xcom L/^[^%]\\begin{/x-'
   getline lin
   parse value lin with '\begin{'name'}'
   return name
*/

--
-- Go to end/beginning of current block
--
defproc LaTeX_end_top_block(back)
   if back then
      beginline; up
      'xcom L/^~([ \t]*%|.*[^\\]%).*(\\begin|\\left)~:a/ex-'
      if rc <> 0 then
         .line = 0
      endif
   else
      beginline; down
      'xcom L/^~([ \t]*%|.*[^\\]%).*(\\end|\\right)~:a/ex+'
      if rc <> 0 then
         .line = .last
      endif
   endif

--
-- Counts the level of nesting in {} blocks.
-- A negative output means that there are more } than
-- {. A positive output means that there are unclosed {.
-- Zero means that braces are balanced.
--
defproc TeX_brace_count(lin)
   n = 0;
   do i = 1 to length(lin)
      c = Substr(lin,i,1)
      if c == '{' then
         n = n + LaTeX_indent_level
      elseif c == '}' then
         n = n - LaTeX_indent_level
      elseif c == '\' then
         if prefix("\{",lin) then
            n = n - LaTeX_indent_level
         elseif prefix("\}",lin) then
            n = n + LaTeX_indent_level
         endif
      endif
   enddo
   return n

--
-- Count indentation change caused by all \left, \right, \begin, and
-- \end commands in the following line.
--
defproc LaTeX_indent_level_count(lin)
   lin = strip_comment(lin)
   i = Pos('\',lin)
   n = 0
   while i > 0 do
      lin = SubStr(lin, i+1)
      i = Pos('\',lin)
      if prefix("left", lin) then
         n = n + LaTeX_left_right_indent_level
      elseif prefix("right", lin) then
         n = n - LaTeX_left_right_indent_level
      elseif prefix("begin", lin) then
         n = n + LaTeX_indent_level
      elseif prefix("end", lin) then
         n = n - LaTeX_indent_level
      elseif i == 1 then
         i = Pos("\",SubStr(lin,2))
      endif
   endwhile
   return n

defproc LaTeX_indent_calc_last
   lastline = .line
   lastcol  = .col

   -- Move back until we find a non empty line (comments excluded)
   -- If no non-empty line is found, then indentation defaults to zero

   beginline; up
   'xcom L/^[ \t]*[^ \t%]/g-'
   if rc /== 0 then
      .lineg = lastline
      .col = lastcol
      return 0
   endif

   getline line
   line = strip_comment(strip_indent(line))

   -- Count the increment in indentation relative to this nonemtpy
   -- line.

   if (.line == 0) then
      n = 0
   elseif prefix("\begin{document}", line) then
      n = current_indentation()
   elseif prefix("\begin{verbatim*}", line) then
      n = 0
   elseif prefix("\end{verbatim*}", line) then
      --
      --FIXME! This regular expression ommits the possible case
      --of a line containing '\%' and '\begin{verbatim*}'
      --
      'xcom L/^[^%]*\\begin{verbatim\*}/x-'
      if rc == 0 then
         n = LaTeX_indent_calc_last()
      else
         n = 0
      endif
   else
      n = current_indentation() + TeX_brace_count(line) + LaTeX_indent_level_count(line)
      if prefix("\end", line) then
         n = n + LaTeX_indent_level
      elseif prefix("\right", line) then
         n = n + LaTeX_left_right_indent_level
      elseif prefix("\item", line) | prefix("\bibitem", line) then
         n = n - LaTeX_item_indent
      endif
   endif
   .lineg = lastline
   .col  = lastcol
   return n

--
-- Returns the amount of indentation corresponding to
-- the line under the cursor.
--
defproc LaTeX_indent_calc
   lastlin = .line
   lastcol = .col

   getline line
   line = strip_indent(line)

   if prefix("\begin{verbatim*}",line) | prefix("\end{verbatim*}",line) then
      n = 0
   elseif prefix("%",line) then
      n = 0
   elseif prefix("\end",line) then
      n = LaTeX_indent_calc_last() - LaTeX_indent_level
   elseif prefix("\right",line) then
      n = LaTeX_indent_calc_last() - LaTeX_left_right_indent_level
   elseif prefix("\item",line) | prefix("\bibitem",line) then
      n = LaTeX_indent_calc_last() + LaTeX_item_indent
   else
      n = LaTeX_indent_calc_last()
   endif
   if n < 0 then
      return 0
   endif
   return n

--
-- Add LaTeX-indent-level indentation in each \begin{ - \end{ block.
-- Lines starting with an item are given an extra indentation of
-- LaTeX-item-indent.
--
-- Final cursor column is undefined.
--
-- Input:	mright = right margin (left is forced to be 1)
--
defproc LaTeX_indent_line3(mright)

   getline text
   text = strip_indent(text)
   loop
      amt = LaTeX_indent_calc()	-- get amount of indentation
      --
      --If the line + indentation exceeds the margins, we only indent
      --a portion and leave the rest to the next try.
      --
      if split_line(mright-amt, line, text) == 0 then
         replaceline indent_str(amt)||line
         leave
      else
         text = strip_indent(text)
         insertline indent_str(amt)||strip(line,'T',' ')
         replaceline text
      endif
   endloop

-- same as above, but tge screen is saved and restored:
defproc LaTeX_indent_line2(mright)
   saved_y = .cursory   -- save cursor y position    ********
   saved_l = .line      -- save line number          ********
   display -1           -- disable screen updating   ********

   getline text
   text = strip_indent(text)
   loop
      amt = LaTeX_indent_calc()	-- get amount of indentation
      --
      --If the line + indentation exceeds the margins, we only indent
      --a portion and leave the rest to the next try.
      --
      if split_line(mright-amt, line, text) == 0 then
         replaceline indent_str(amt)||line
         leave
      else
         text = strip_indent(text)
         insertline indent_str(amt)||strip(line,'T',' ')
         replaceline text
      endif
   endloop
   .cursory = saved_y   -- restore cursor position
   .line = saved_l      -- restore line number
   display 1            -- enable screen updating

-- the third variant:  the second arg controls
-- whether the cursor should advance
defproc LaTeX_indent_line(mright,advance)
   -- 
   -- If this is called in conjunction with the <enter> key, then
   -- advance =1, els advance=0
   --
   -- First, get a stripped line and store the column where the cursor
   -- is on, but relative to the indentation. That is, without caring for
   -- leading tabs and spaces.
   --
   saved_y = .cursory        -- save cursor y position
   saved_l = .line           -- save line number
   display -1                -- disable screen updating
   getline text
   text = strip_indent_col(text, .col, oldcolumn)
   if oldcolumn < 0 then
      -- Pointer was in the indentation zone. Thus, we
      -- prepare it to be on the first non-blank char
      oldcolumn = 0
   endif

   --
   --If the line + indentation exceeds the margins, we only indent
   --a portion and leave the rest to the next try.
   --
   amt = LaTeX_indent_calc()	-- get amount of indentation

   sp = split_line(mright-amt, line, text)
   if sp == 0 then
      replaceline indent_str(amt)||line
      .col = oldcolumn + amt
   else
      text = strip_indent_col(text, oldcolumn - Length(line), aux)
      insertline indent_str(amt)||strip(line,'T',' ')
      replaceline text
      if aux >= 0 then
         .col = aux + sp - 1
         call LaTeX_indent_line(mright)
      else
         oldline = .line - 1
         call LaTeX_indent_line2(mright) -- better ...3 ?
         .lineg = oldline
         .col = oldcolumn + current_indentation() + sp - 1
      endif
   endif
   .cursory = saved_y      -- restore cursor position
   .line = saved_l+advance -- restore line number
   display 1               -- enable screen updating

/*
--
-- First, get a stripped line and store the column where the cursor
-- is on, but relative to the indentation. That is, without caring for
-- leading tabs and spaces.
--
   getline text
   updatepos = 1
   oldline = .line
   text = strip_indent_col(text, .col, oldcolumn)
   if oldcolumn < 0 then
      -- Pointer was in the indentation zone. Thus, we
      -- prepare it to be on the first non-blank char
      oldcolumn = 0
      updatepos = 0
   endif

   loop
      --
      --If the line + indentation exceeds the margins, we only indent
      --a portion and leave the rest to the next try.
      --
      amt = LaTeX_indent_calc()	-- get amount of indentation

      if split_line(mright-amt, line, text) == 0 then
         if updatepos == 0 then
            replaceline indent_str(amt)||line
         else
            line = strip_indent_col(line, oldcolumn, aux)
            if aux < 0 then
               oldcolumn = 0
            else
               oldcolumn = aux
            endif
            replaceline indent_str(amt)||line
         endif
         leave
      else
         if updatepos == 0 then
            text = strip_indent(text)
         else
            text = strip_indent_col(text, oldcolumn, aux)
            aux = aux - Length(line)
            if aux < 0 then
               updatepos = 0
            else
               oldcolumn = aux
               oldline = oldline + 1
            endif
         endif
         insertline indent_str(amt)||strip(line,'T',' ')
         replaceline text
      endif
   endloop
   .lineg = oldline
   .col = oldcolumn + current_indentation()
*/

defproc LaTeX_joinlines_unindented()
   getline line1
   getline line2, .line + 1
   deleteline .line + 1
   replaceline line1' 'strip_indent(line2)

defproc LaTeX_fill_region_as_paragraph(mright)
   saved_y = .cursory   -- save cursor y position    ********
   saved_l = .line      -- save line number          ********
   display -1           -- disable screen updating   ********
   if MarkType() == '' then
      beginline
      'xcom L/^[ \t]*[^ \t%]/g'
      markline
      .lineg = .lineg + 1
      beginline
      'xcom L/'LaTeX_paragraph_start'/x'
      if rc /== 0 then
         .lineg = .last
      else
         .lineg = .lineg - 1
      endif
      markline
   endif
   getmark firstline, lastline, foo, foo, markfileid
   getfileid thisfileid
   if markfileid <> thisfileid then
      call WinMessageBox('LaTeX','No mark in this file')
      stop
   endif
   call pbegin_mark()
   unmark
   for i = 0 to (lastline - firstline)
      if i > 0 then
         getline line
         if LeftStr(strip_indent(line),1) == '%' then
            down
         else
            call LaTeX_joinlines_unindented()
         endif
      endif
      call LaTeX_indent_line3(mright) -- don't save and restore here!
   endfor
   .cursory = saved_y   -- restore cursor position
   .line = saved_l      -- restore line number
   display 1            -- enable screen updating
-- End up in the end of the paragraph
--   endline

--
-- The INDENT command. Indents the current line, leaving the
-- cursor in the right place
--
defc indent
   universal TeX_Margin
   if MarkType() == '' then
      beginline
      'xcom L/^[ \t]*[^ \t%]/g'
      markline; markline
      .lineg = .lineg + 1;
      beginline
      'xcom L/'LaTeX_paragraph_start'/x'
      if rc /== 0 then
         .line = .last
      else
         up
      endif
      markline
   endif
   getmark firstline, lastline, foo, foo, markfileid
   getfileid thisfileid
   if markfileid <> thisfileid then
      call WinMessageBox('LaTeX','No mark in this file')
      stop
   endif
   call pbegin_mark()
   unmark
   for i = 0 to (lastline - firstline)
      if i > 0 then
         getline line
         .lineg = .lineg + 1
      endif
      call LaTeX_indent_line2(TeX_Margin)
   endfor

--
-- FILLPAR indents a marked paragaph or finds the end of the
-- current one and indents and reflows it.
--
defc fillpar
   universal TeX_Margin
   call LaTeX_fill_region_as_paragraph(TeX_Margin)

/* Not required in conjunction with the TeX Front End:
--
-- TEXMA, TEXMARGINS sets the line length for TeX indentation.
-- It resets MARGINS (the EPM ones) to '1 1599 1' so that
-- they don't interfere with TeX indentation.
--
defc texma
   universal TeX_Margin
   parse arg l r p
   if l = '' then
      TeX_Margin = 72
   elseif r = '' then
      TeX_Margin = l
   else
      TeX_Margin = r
   endif
   'xcom ma 1 1599 1'
   sayerror 'New TeX margin 'TeX_Margin

defc texmargins
   parse arg l r p
   'texma' l r p
*/

/*
--
-- EOP = find End Of Paragraph
--
defc eop
   beginline; down
   'xcom L/'LaTeX_paragraph_start'/x'
   up
*/

/* changed 1998-03-08: defc texma and defc texmargins commented out */

/* changed 1998-03-23: prevent text from 'jumping' (WaS)            */

